<?php
use CRM_Updaterecur_ExtensionUtil as E;

/**
 * Collection of upgrade steps.
 */
class CRM_Updaterecur_Upgrader extends CRM_Extension_Upgrader_Base {

  // By convention, functions that look like "function upgrade_NNNN()" are
  // upgrade tasks. They are executed in order (like Drupal's hook_update_N).

  /**
   * Example: Run an external SQL script when the module is installed.
   *
  public function install() {
    $this->executeSqlFile('sql/myinstall.sql');
  }

  /**
   * Example: Work with entities usually not available during the install step.
   *
   * This method can be used for any post-install tasks. For example, if a step
   * of your installation depends on accessing an entity that is itself
   * created during the installation (e.g., a setting or a managed entity), do
   * so here to avoid order of operation problems.
   */
   public function postInstall() {
     // In install context we apparently don't have this->ctx, but we use it in ensureDataStructuresExist
     if (empty($this->ctx->log)) {
       $this->ctx = $this->ctx ?? new StdClass();
       $this->ctx->log = \Civi::log();
     }
     CRM_Core_Transaction::create()->run([$this, 'ensureDataStructuresExist']);
   }

   /**
    */
  public function ensureDataStructuresExist() {

    //
    // Create activity type(s).
    //
    $activityTypeNamesToValues = [];
    foreach ([
      [
        'name' => 'updaterecur_change',
        'label' => 'Recurring Contribution change',
        'description' => 'Logs details of an (attempted, scheduled or completed) update to an existing recurring contribution/payment subscription, as part of a bulk operation.',
        'icon' => 'fa-money',
      ],
    ] as $activityType) {
      $activityTypeNamesToValues[$activityType['name']] = Civi\Api4\OptionValue::get()
        ->setCheckPermissions(FALSE)
        ->addWhere('option_group_id:name', '=', 'activity_type')
        ->addWhere('name', '=', $activityType['name'])
        ->execute()
        ->first()['value'] ?? NULL;
      if (empty($activityTypeNamesToValues[$activityType['name']])) {
        // Not found, create it now.
        $activityTypeNamesToValues[$activityType['name']] = \Civi\Api4\OptionValue::create()
          ->setCheckPermissions(FALSE)
          ->addValue('option_group_id:name', 'activity_type')
          ->addValue('label', $activityType['name'])
          ->addValue('name', $activityType['name'])
          ->addValue('description', $activityType['description'])
          ->addValue('is_active', 1)
          // ->addValue('color', $colour)
          // ->addValue('grouping', $tpl['grouping'])
          ->execute()->first()['value'];
      }
    }

    //
    // Create custom field group for the activity type.
    //
    $baseParams = [
      'name'       => 'updaterecur_activity_fields',
      'table_name' => 'civicrm_updaterecur_activity_fields',
    ];
    $allParams = [
      "title"                       => "Recurring Contribution Changes",
      "extends"                     => "Activity",
      "extends_entity_column_value" => [ $activityTypeNamesToValues['updaterecur_change'] ],
      "style"                       => "Inline",
      "collapse_display"            => "0",
      "is_active"                   => "1",
      "collapse_adv_display"        => "0",
      "is_reserved"                 => "1", // ???
      "is_public"                   => "0",
    ];
    $customGroupIDChange = $this->createOrUpdate('CustomGroup', $baseParams, $allParams);

    // Create fake FK to ContributionRecur
    $baseParams = [
      'custom_group_id' => $customGroupIDChange,
      'name'            => "updaterecur_recur_id",
    ];
    $allParams = [
      'column_name'   => "contribution_recur_id",
      'label'         => "Recurring Contribution ID",
      'data_type'     => "Int",
      'html_type'     => "Text",
      "is_view"       => "1",
      'is_searchable' => 1,
      'is_required'   => 1,
    ];
    $this->createOrUpdate('CustomField', $baseParams, $allParams);

    // Field to store old amount.
    $baseParams = [
      'custom_group_id' => $customGroupIDChange,
      'name'            => "updaterecur_old_amount",
    ];
    $allParams = [
      'column_name'   => "old_amount",
      'label'         => "Old amount",
      'data_type'     => "Float",
      'html_type'     => "Text",
      "is_view"       => 1,
      'is_searchable' => 1,
      'is_required'   => 1,
    ];
    $this->createOrUpdate('CustomField', $baseParams, $allParams);

    // Field to store new amount.
    $baseParams = [
      'custom_group_id' => $customGroupIDChange,
      'name'            => "updaterecur_new_amount",
    ];
    $allParams = [
      'column_name'   => "new_amount",
      'label'         => "New amount",
      'data_type'     => "Float",
      'html_type'     => "Text",
      "is_view"       => 1,
      'is_searchable' => 1,
      'is_required'   => 1,
    ];
    $this->createOrUpdate('CustomField', $baseParams, $allParams);

  }
  /**
   * Helper function
   *
   * @param string $entity e.g. Contact
   * @param array $baseParams used to find the entity
   * @param array $allParams used to update details of the entity. Added to $baseParams for new entities.
   * @param bool $update. If entity found, should we still do a create call to update it?
   *
   * @return int ID of thing created.
   */
  protected function createOrUpdate($entity, $baseParams, $allParams, $update=TRUE) {
    $results = civicrm_api3($entity, 'get', $baseParams);

    $logPrefix = "createOrUpdate $entity with " . json_encode($baseParams);

    $params = $allParams;
    $debug_details = '';
    $count = $results['count'];
    if ($count === 1) {
      if (!$update) {
        $this->ctx->log->info("$logPrefix: found entity $results[id], no change");
        return $results['id'];
      }
      $this->ctx->log->info("$logPrefix: found entity $results[id], will call update");
      $params['id'] = $results['id'];
      $debug_details = 'update';
    }
    elseif ($count === 0) {
      $this->ctx->log->info("$logPrefix: entity NOT found, will create");
      // Create.
      $params += $baseParams;
      $debug_details = 'create';
    }
    elseif ($count > 1) {
      throw new Exception("Found multiple $entity entities matching: " . json_encode($baseParams));
    }
    //print "$debug_details entity.create " . json_encode($params, JSON_PRETTY_PRINT) . "\n";
    $id = (int) civicrm_api3($entity, 'create', $params)['id'] ?? NULL;
    if (!$id) {
      throw new \Exception("Failed to $debug_details '$entity' entity with: " . json_encode($params));
    }
    $this->ctx->log->info("$logPrefix: created entity $id");
    return $id;
  }

  /**
   * Example: Work with entities usually not available during the install step.
   *
   * This method can be used for any post-install tasks. For example, if a step
   * of your installation depends on accessing an entity that is itself
   * created during the installation (e.g., a setting or a managed entity), do
   * so here to avoid order of operation problems.
   */
  // public function postInstall() {
  //  $customFieldId = civicrm_api3('CustomField', 'getvalue', array(
  //    'return' => array("id"),
  //    'name' => "customFieldCreatedViaManagedHook",
  //  ));
  //  civicrm_api3('Setting', 'create', array(
  //    'myWeirdFieldSetting' => array('id' => $customFieldId, 'weirdness' => 1),
  //  ));
  // }

  /**
   * Example: Run an external SQL script when the module is uninstalled.
   */
  // public function uninstall() {
  //  $this->executeSqlFile('sql/myuninstall.sql');
  // }

  /**
   * Example: Run a simple query when a module is enabled.
   */
  // public function enable() {
  //  CRM_Core_DAO::executeQuery('UPDATE foo SET is_active = 1 WHERE bar = "whiz"');
  // }

  /**
   * Example: Run a simple query when a module is disabled.
   */
  // public function disable() {
  //   CRM_Core_DAO::executeQuery('UPDATE foo SET is_active = 0 WHERE bar = "whiz"');
  // }

  /**
   * Example: Run a couple simple queries.
   *
   * @return TRUE on success
   * @throws Exception
   */
  // public function upgrade_4200() {
  //   $this->ctx->log->info('Applying update 4200');
  //   CRM_Core_DAO::executeQuery('UPDATE foo SET bar = "whiz"');
  //   CRM_Core_DAO::executeQuery('DELETE FROM bang WHERE willy = wonka(2)');
  //   return TRUE;
  // }


  /**
   * Example: Run an external SQL script.
   *
   * @return TRUE on success
   * @throws Exception
   */
  // public function upgrade_4201() {
  //   $this->ctx->log->info('Applying update 4201');
  //   // this path is relative to the extension base dir
  //   $this->executeSqlFile('sql/upgrade_4201.sql');
  //   return TRUE;
  // }


  /**
   * Example: Run a slow upgrade process by breaking it up into smaller chunk.
   *
   * @return TRUE on success
   * @throws Exception
   */
  // public function upgrade_4202() {
  //   $this->ctx->log->info('Planning update 4202'); // PEAR Log interface

  //   $this->addTask(E::ts('Process first step'), 'processPart1', $arg1, $arg2);
  //   $this->addTask(E::ts('Process second step'), 'processPart2', $arg3, $arg4);
  //   $this->addTask(E::ts('Process second step'), 'processPart3', $arg5);
  //   return TRUE;
  // }
  // public function processPart1($arg1, $arg2) { sleep(10); return TRUE; }
  // public function processPart2($arg3, $arg4) { sleep(10); return TRUE; }
  // public function processPart3($arg5) { sleep(10); return TRUE; }

  /**
   * Example: Run an upgrade with a query that touches many (potentially
   * millions) of records by breaking it up into smaller chunks.
   *
   * @return TRUE on success
   * @throws Exception
   */
  // public function upgrade_4203() {
  //   $this->ctx->log->info('Planning update 4203'); // PEAR Log interface

  //   $minId = CRM_Core_DAO::singleValueQuery('SELECT coalesce(min(id),0) FROM civicrm_contribution');
  //   $maxId = CRM_Core_DAO::singleValueQuery('SELECT coalesce(max(id),0) FROM civicrm_contribution');
  //   for ($startId = $minId; $startId <= $maxId; $startId += self::BATCH_SIZE) {
  //     $endId = $startId + self::BATCH_SIZE - 1;
  //     $title = E::ts('Upgrade Batch (%1 => %2)', array(
  //       1 => $startId,
  //       2 => $endId,
  //     ));
  //     $sql = '
  //       UPDATE civicrm_contribution SET foobar = whiz(wonky()+wanker)
  //       WHERE id BETWEEN %1 and %2
  //     ';
  //     $params = array(
  //       1 => array($startId, 'Integer'),
  //       2 => array($endId, 'Integer'),
  //     );
  //     $this->addTask($title, 'executeSql', $sql, $params);
  //   }
  //   return TRUE;
  // }

}
