<?php
namespace Civi\api4\Action\ContributionRecur;

use CRM_Updaterecur_ExtensionUtil as E;

/**
 * This API Action only creates activities for the given ContributionRecur IDs.
 *
 * Those will be processed later by a Job.
 */
class ChangeSubscriptions extends \Civi\Api4\Generic\AbstractAction {

  /**
   * We need to support the Where clause for crmSearchBatchRunner
   *
   * @var array
   */
  protected $where;

  /**
   * New amount
   *
   * @var string
   * @required
   */
  protected $newAmount;

  /**
   * Reason (for the change). ex. "General membership fee increase"
   *
   * @var string
   * @required
   */
  protected $reason;

  /**
   * This only exists here so that we can provide a different result class by
   * way of the annotation below.
   *
   * @return \Civi\Api4\ResultWithMetadata
   */
  public function execute() {
    return parent::execute();
  }

  /**
   *
   * Note that the result class is that of the annotation below, not the hint
   * in the method (which must match the parent class)
   *
   * @var \Civi\Api4\ResultWithMetadata $result
   */
  public function _run(\Civi\Api4\Generic\Result $result) {

    $whoIsDoingThis = \CRM_Core_Session::singleton()->getLoggedInContactID();
    if (!$whoIsDoingThis) {
      $whoIsDoingThis = \CRM_Core_BAO_Domain::getDomain()->contact_id;
    }

    // We have limited, specific filters available. Let's make that clear.
    if (!is_array($this->where)
      || count($this->where) !== 1
      || !is_array($this->where[0])
      || count($this->where[0]) !== 3
      || $this->where[0][0] !== 'id'
      || $this->where[0][1] !== 'IN'
      || !is_array($this->where[0][2])
      || empty($this->where[0][2])
    ) {
      throw new \API_Exception("ContributionRecur.ChangeSubscriptions requires one 'where' clause which is ['id', 'IN', [ID,...]]. Got: " . json_encode($this->where));
    }

    // Load the recurs.
    $recurs = \Civi\Api4\ContributionRecur::get(FALSE)
      ->addSelect('contact_id', 'id', 'amount', 'processor_id', 'contribution_status_id:label', 'payment_processor_id', 'financial_type_id:label', 'currency')
      ->setWhere($this->where)
      ->execute()->indexBy('id');

    // Load any Scheduled activities relating to these recurs. If we find any, stop right now.
    $activities = \Civi\Api4\Activity::get(FALSE)
      ->addSelect('updaterecur_activity_fields.updaterecur_recur_id')
      ->addWhere('activity_type_id:name', '=', 'updaterecur_change')
      ->addWhere('status_id:name', '=', 'Scheduled')
      ->addWhere('updaterecur_activity_fields.updaterecur_recur_id', 'IN', $recurs->column('id'))
      ->execute();
    if (count($activities->getArrayCopy())) {
      throw new \API_Exception("There are changes pending for the following ContributionRecur records with the following IDs: "
        . json_encode($activities->column('updaterecur_activity_fields.updaterecur_recur_id')));
    }

    // We want them ALL to work, or none of them to.
    \CRM_Core_Transaction::create()->run(function() use ($result, $recurs, $whoIsDoingThis) {

      $result->metadata = [
        'scheduled' => 0,
        'skipped' => 0,
      ];
      $activities = [];
      $newAmount = (float) $this->newAmount;
      $now = date('Y-m-d H:i:s');
      $sillyLogMessage = E::ts('This activity was created in a Cancelled state because the requested new amount is the same as the existing amount.');

      foreach ($recurs as $recur) {
        $currency = $recur['currency'];
        $oldAmount = (float) $recur['amount'];
        $activity = [
          'updaterecur_activity_fields.updaterecur_recur_id' => $recur['id'],
          'updaterecur_activity_fields.updaterecur_old_amount' => $recur['amount'],
          'target_contact_id' => $recur['contact_id'],
          'details' => "<p>$now " . E::ts(
              ($newAmount > $oldAmount)
              ? "Increase amount from %1 to %2 %3."
              : (
                ($newAmount < $oldAmount)
                ? "Decrease amount from %1 to %2 %3."
                : "Change amount from %1 to %2 %3."
              ),
              [1 => $oldAmount, 2 => $newAmount, 3 => $currency]
          ) . ' ',
        ];
        if ($newAmount === $oldAmount) {
          // Erm...let's record this mistake.
          $activity += [
            'activity_date_time' => date('Y-m-d H:i:s'),
            'status_id:name' => 'Cancelled',
          ];
          $activity['details'] .= $sillyLogMessage;
          $result->metadata['skipped']++;
        }
        else {
          $activity['details'] .= E::ts('Activity Scheduled for processing from 30 mins’ time.');
          $result->metadata['scheduled']++;
        }
        $activity['details'] .= "</p>\n";
        $activities[] = $activity;
      }
      // We create the activities with a datetime that's 30mins in the future.
      // This means if you make a mistake you have 30mins to delete those activities...
      $createResult = \Civi\Api4\Activity::save(FALSE)
        ->setDefaults([
          'subject'               => $this->reason,
          'activity_type_id:name' => 'updaterecur_change',
          'status_id:name'        => 'Scheduled',
          'activity_date_time'    => date('Y-m-d H:i:s', time() + 60 * 30),
          'source_contact_id'     => $whoIsDoingThis,
          'updaterecur_activity_fields.updaterecur_new_amount' => $this->newAmount,
        ])
        ->setRecords($activities)
        ->execute();

      $result->exchangeArray($createResult->getArrayCopy());
    });

    return $result;
  }

}
