<?php

namespace Civi\Inlay;

use CRM_Updaterecur_ExtensionUtil as E;
use Civi\UpdaterecurWorker;
use Civi\Updaterecur\Utils;
use Brick\Money\Money;
use Civi\Api4\Activity;
use Civi\Inlay\Type as InlayType;
use Civi\Api4\Contact;
use Civi\Api4\ContributionRecur;
use CRM_Core_Config;
use Civi\Payment\Exception\PaymentProcessorException;
use Civi;

class UpdateGiving extends InlayType {

  const EVENT_AFTER_PREFILL = 'civi.inlayupdategiving.afterprefill';
  const EVENT_CHANGE_REQUEST_ALTER = 'civi.inlayupdategiving.changerequestalter';
  const EVENT_AFTER_CHANGE = 'civi.inlayupdategiving.afterchange';

  const DIDNT_WORK = "Sorry, we were unable to update your giving. Don’t worry, this can happen for all sorts of reasons. Please get in touch so we can sort it out.";

  public static $typeName = 'Update Giving';
  public static $machineName = 'UpdateGiving';

  public static $defaultConfig = [
  // replaces frontend_title v1.5
    'publicTitle'                    => 'Donate',
  // replaces intro_text v1.5
    'publicIntroHTML'                => '',
    'thanksHTML' => 'Thank you!',
    'thanksMessageTemplateID' => NULL,
    'thanksMessageFrom' => NULL,
    'thanksMessageBCC' => NULL,
    'donatePageURL' => NULL,
    'reason' => '',
    'testMode' => FALSE,
    'css' => '',
  ];

  /**
   * Generates data to be served with the Javascript application code bundle.
   *
   * @return array
   */
  public function getInitData(): array {
    // \Civi::log()->debug("inlayupdategiving getInitData config is " . json_encode($this->config, JSON_PRETTY_PRINT));

    $copyFromConfig = [
      'publicTitle',
      'publicIntroHTML',
      'thanksHTML',
      'donatePageURL',
      'testMode',
      'css',
    ];
    $data = array_intersect_key($this->config, array_flip($copyFromConfig));
    $data['init'] = 'inlayUpdateGivingInit';

    return $data;
  }

  /**
   * Process a request
   *
   * @param \Civi\Inlay\Request $request
   * @return array
   *
   * @throws \Civi\Inlay\ApiException;
   */
  public function processRequest(ApiRequest $request): array {

    $input = $request->getBody();
    switch ($input['need']) {
      case 'prefillData':
        return $this->prefillData($input);

      case 'change':
        return $this->handleChangeRequest($input);

      default:
        throw new \Exception("error unrecognised need");
    }
  }

  public function prefillData(array $requestBody): array {
    $output = $this->loadAllData($requestBody);
    // Strip out data we don't need publicly.
    $output = array_intersect_key($output, array_flip([
      'descriptions',
      'suggested',
      'updateDirectly',
      'currentAmount',
      'first_name',
      'last_name',
    ]));
    return $output;
  }

  public function loadAllData(array $requestBody): array {
    $output = [];
    $contactID = $requestBody['ctID'] ?? 0;
    $checksum = $requestBody['cs'] ?? '';

    if (!\CRM_Contact_BAO_Contact_Utils::validChecksum($contactID, $checksum)) {
      Civi::log()->warning("inlayupdategiving: prefill: Invalid checksum received for ct{$contactID}");
      return [];
    }
    Civi::log()->info("inlayupdategiving: prefill: Valid checksum received for ct{$contactID}");

    $contact = Contact::get(FALSE)
      ->addWhere('id', '=', $contactID)
      ->addSelect('first_name', 'last_name', 'email_primary.email')
      ->execute()->first();
    if ($contact) {
      $output['first_name'] = $contact['first_name'] ?? '';
      $output['last_name'] = $contact['last_name'] ?? '';
      $output['email'] = $contact['email_primary.email'] ?? '';
    }

    // Get descriptions of giving.
    $output += $this->getGivingDetails($contactID);

    return $output;
  }

  /**
   * Handles the 2nd ajax call, intending to change a subscription.
   */
  public function handleChangeRequest(array $requestBody): array {
    // Load data as before, should work.
    $loadedData = $this->loadAllData($requestBody);
    if (!$loadedData || count($loadedData['crs']) !== 1) {
      Civi::log()->warning("inlayupdategiving error/misuse (public error code IUG2): handleChangeRequest called but we don't have a single cr.", compact('requestBody', 'loadedData') + [
        'inlay' => $this->getName(),
        'testMode' => $this->config['testMode'],
      ]);
      return ['error' => 'Sorry, something went wrong. Please let us know (code IUG2)'];
    }
    $cr = $loadedData['crs'][0];
    if ($this->config['testMode'] && empty($cr['is_test'])) {
      return ['error' => 'In test mode, refusing to update a real subscription!'];
    }
    // Validate input.
    $newAmountPennies = Money::of($requestBody['newAmount'], $cr['currency'])->getMinorAmount()->toInt();
    $oldAmountPennies = Money::of($cr['amount'], $cr['currency'])->getMinorAmount()->toInt();
    if ($newAmountPennies < 100) {
      return ['error' => 'Amount is below the minimum.'];
    }
    elseif (abs($newAmountPennies - $oldAmountPennies) < 50) {
      return ['error' => 'Sorry, we can’t process amount changes that small.'];
    }
    // Safety check: don't allow more than 10× increases if the result is over £50
    if (($newAmountPennies / $oldAmountPennies) > 10 && $newAmountPennies >= 5000) {
      $symbol = ['GBP' => '£', 'USD' => '$', 'CAD' => '$', 'EUR' => '€'][$cr['currency']] ?? $cr['currency'];
      $newAmountPounds = (string) Money::of($requestBody['newAmount'], $cr['currency'])->getAmount();
      return [
        'error' => "This looks like an exceedingly big increase in giving. "
        . "We would be very grateful for this, but for your safety please contact us to confirm "
        . "that you want to regularly give $symbol$newAmountPounds.",
      ];
    }
    // Allow local code to do its own checks. Handlers can set 'error' to stop things moving forward.
    // They can set notifyOnSuccess FALSE to prevent our own success notification routine.
    $return = ['notifyOnSuccess' => TRUE];
    $event = \Civi\Core\Event\GenericHookEvent::create(['data' => &$return, 'loadedData' => $loadedData, 'requestBody' => $requestBody, 'inlay' => $this]);
    \Civi::dispatcher()->dispatch(self::EVENT_CHANGE_REQUEST_ALTER, $event);
    if (!empty($return['error'])) {
      return $return;
    }

    try {
      $this->attemptSubscriptionChange($return, $loadedData, $requestBody);
      $event = \Civi\Core\Event\GenericHookEvent::create(['data' => &$return, 'loadedData' => $loadedData, 'requestBody' => $requestBody, 'inlay' => $this]);
      \Civi::dispatcher()->dispatch(self::EVENT_AFTER_CHANGE, $event);
      Civi::log()->debug("UpdateGiving: ", [
        'return' => $return,
        'loadedData' => $loadedData,
      ]);
      if (!empty($return['error'])) {
        return $return;
      }

      if ($return['notifyOnSuccess']) {
        $this->sendThankYouEmail($return, $loadedData, $requestBody);
      }
    }
    catch (\Exception $e) {
      Civi::log()->error("UpdateGiving: FAILED trying to change subscription - general exception", [
        'message'               => $e->getMessage(),
        'type'                  => get_class($e),
        'contact_id'            => $loadedData['contactID'],
        'contribution_recur_id' => $cr['contribution_recur_id'],
        'amount'                => $requestBody['amount'],
      ]);
      $return = ['error' => 'Sorry, something went wrong. Please contact us. (Code IUG1)'];
    }

    return $return;
  }

  /**
   */
  public function getGivingDetails(int $contactID): array {
    $return = Utils::getGivingDetails($contactID, $this->config['testMode']);
    $currentAmount = $suggested = '';
    $updateDirectly = FALSE;
    if (count($return['crs']) === 1) {
      $cr = reset($return['crs']);
      $updateDirectly = $cr['payment_processor_id.payment_processor_type_id:name'] === 'GoCardless';
      // Although we can only change it directly for GC, it's still useful to have the suggestion to
      // pass on to the new-payment form.
      $suggested = static::suggestedUpgrade($cr['amount']);
      $currentAmount = $cr['amount'];
    }
    return $return + compact('suggested', 'currentAmount', 'updateDirectly');
  }

  public function updateRecurAmount(int $crID, string $newAmount) {
    // Now Update CiviCRM's CR. We also need to update the template...
    civicrm_api3('ContributionRecur', 'create', [
      'id'     => $crID,
      'amount' => $newAmount,
    ]);
  }

  /**
   * Get the Javascript app script.
   *
   * This will be bundled with getInitData() and some other helpers into a file
   * that will be sourced by the client website.
   *
   * @return string Content of a Javascript file.
   */
  public function getExternalScript(): string {
    return file_get_contents(E::path('js/inlayupdategiving.js'));
  }

  /**
   * Return the configured currency, or CiviCRM's default configured currency.
   */
  public function getCurrency() {
    return ($this->config['currency'] ?? '') ?: CRM_Core_Config::singleton()->defaultCurrency;
  }

  /**
   * Get the URL template for editing instances of this inlay type.
   */
  public static function getInstanceEditURLTemplate(): string {
    // Construct the default URL template.
    // Note: because of the way CRM.url works, you MUST put a ? before the #
    return 'civicrm/a?#/inlays/' . static::getTypeMachineName() . '/{id}';
  }

  /**
   * This tries to contact the payment processor to change the subscription.
   *
   * If unsuccessful, it should populate $return['error']
   * If successful, it should populate $return['summary'] with a plain text string and $return['success'] with true.
   */
  public function attemptSubscriptionChange(array &$return, array &$loadedData, array $requestBody): void {
    $cr = $loadedData['crs'][0];

    try {
      $activity = ContributionRecur::changeSubscriptions(FALSE)
        ->setWhere([['id', 'IN', [$cr['id']]]])
        ->setNewAmount($requestBody['newAmount'])
        ->setReason($this->config['reason'] ?: ('Update Giving inlay: ' . $this->getName()))
        ->execute()->first();
      // Add some context and force the date to the past.
      $this->updateActivityDetails($activity,
         "<p>Immediate processing attempt by Update Giving Inlay <em>" . htmlspecialchars($this->getName())
        . "</em> at <a href=\"" . htmlspecialchars($requestBody['sourceURL']) . "\" rel='noopener noreferrer' target=_blank >"
        . htmlspecialchars($requestBody['sourceURL']) . "</a></p>\n",
        ['activity_date_time' => date('YmdHis', time() - 30)]);

      $u = new UpdaterecurWorker(30);
      $success = $u->processActivity($activity);
      $activity = Activity::get(FALSE)->addSelect('*', 'custom.*')->addWhere('id', '=', $activity['id'])->execute()->first();

      // Make the activity accessible to other processes.
      $loadedData['activity'] = $activity;

      if ($success) {
        $return['success'] = TRUE;
        // Reload the cr.
        $cr = ContributionRecur::get(FALSE)
          ->addWhere('id', '=', $cr['id'])
          ->addSelect('amount', 'currency',
            'frequency_unit:label', /*month*/
            'frequency_interval',
            'payment_processor_id',
            'payment_processor_id.payment_processor_type_id:name',
            'payment_instrument_id:label',
            'financial_type_id:name',
            'financial_type_id:label',
          )
          ->execute()->first();
        $return['summary'] = 'Successfully changed, you are now giving ' . Utils::summariseCrAmountAndInterval($cr);
      }
      else {
        $return['error'] = self::DIDNT_WORK;
      }
    }
    catch (PaymentProcessorException $e) {
      $return['error'] = self::DIDNT_WORK;
      Civi::log()->error("UpdateGiving: FAILED trying to change subscription", [
        'message'               => $e->getMessage(),
        'contact_id'            => $loadedData['contactID'],
        'contribution_recur_id' => $cr['id'],
        'newAmount'             => $requestBody['newAmount'],
      ]);
    }
  }

  /**
   * Send thank you email, if configured to do so.
   */
  public function sendThankYouEmail(array &$return, array $loadedData, array $requestBody): void {
    if (!$this->config['thanksMessageTemplateID']) {
      Civi::log()->debug("UpdateGiving: skipping thanks, not configured");
      $this->updateActivityDetails($loadedData['activity'], "<p>Not sending thank you email, as none configured</p>");
      return;
    }
    $contactID = $loadedData['contactID'];

    if ($this->config['thanksMessageFrom']) {
      // Try to load the specified entry.
      $combined = \Civi\Api4\OptionValue::get(FALSE)
        ->setCheckPermissions(FALSE)
        ->addSelect('label')
        ->addWhere('option_group_id:name', '=', 'from_email_address')
        ->addWhere('is_active', '=', TRUE)
        ->addWhere('value', '=', (int) $this->config['thanksMessageFrom'])
        ->addOrderBy('is_default', 'DESC')
        ->execute()->first()['label'] ?? NULL;
    }
    if (empty($combined)) {
      // Nothing configured, or loading the configured address failed.
      // Load default.
      $combined = civicrm_api3('OptionValue', 'getvalue', ['return' => "label", 'option_group_id' => "from_email_address", 'is_default' => 1]);
    }

    $params = [
      'id'         => $this->config['thanksMessageTemplateID'],
      'to_email'   => $loadedData['email'],
      'contact_id' => $contactID,
      'from'       => $combined,
      'template_params' =>
      // {$summary} in templates
      [
        'summary' => $return['summary'],
      ],
    ];
    if (!empty($this->config['thanksMessageBCC'])) {
      $params['bcc'] = $this->config['thanksMessageBCC'];
    }
    Civi::log()->debug("UpdateGiving: MessageTemplate.send params", $params);
    try {
      civicrm_api3('MessageTemplate', 'send', $params);
      Civi::log()->info("inlayupdategiving sent thank you email to contact $contactID at {$loadedData['email_primary.email']}");
      $return['thanksSent'] = TRUE;
      $this->updateActivityDetails($loadedData['activity'], "<p>Thank you email sent to " . htmlspecialchars($loadedData['email']) . "</p>");
    }
    catch (\Exception $e) {
      Civi::log()->error("inlayupdategiving exception while trying to send thank you email to contact $contactID at {$loadedData['email_primary.email']}", [
        'exception' => $e->getMessage(),
        'Email.send api params' => $params,
      ]);
      $return['thanksSent'] = FALSE;
      $this->updateActivityDetails($loadedData['activity'], "<p>Thank you email failed to send to " . htmlspecialchars($loadedData['email']) . " See logs.</p>");
    }
  }

  protected function updateActivityDetails(array &$activity, string $appendHTML, array $otherValues = []) {
    if (empty($activity['id'])) {
      throw new \InvalidArgumentException("updateActivityDetails called with activity without id");
    }
    if (empty($activity['details'])) {
      $activity = Activity::get(FALSE)->addSelect('*', 'custom.*')->addWhere('id', '=', $activity['id'])->execute()->first();
    }
    $api = Activity::update(FALSE)
      ->addWhere('id', '=', $activity['id']);
    if ($otherValues) {
      $api->setValues($otherValues);
    }
    $api->addValue('details', $activity['details'] . $appendHTML)
      ->execute();
    // Reload
    $activity = Activity::get(FALSE)->addSelect('*', 'custom.*')->addWhere('id', '=', $activity['id'])->execute()->first();
  }

  public static function suggestedUpgrade(float $currentAmount): string {
    if ($currentAmount < 7) {
      $suggested = round($currentAmount + 3);
    }
    elseif ($currentAmount < 10) {
      $suggested = round($currentAmount + 4);
    }
    elseif ($currentAmount < 20) {
      $suggested = round($currentAmount + 5);
    }
    elseif ($currentAmount <= 50) {
      $suggested = round($currentAmount + 10);
    }
    else {
      $suggested = round(ceil($currentAmount * 1.20 / 5) * 5);
    }
    return $suggested;
  }

}

// cSpell:ignore trxn getsingle gocardless cardless inlayupdategiving postsuccess directdebit oneoff paymentintent
